@@ -50,13 +50,17 @@ from ocsmesh.hfun.base import BaseHfun
 from ocsmesh.hfun.raster import HfunRaster
 from ocsmesh.hfun.mesh import HfunMesh
 from ocsmesh.mesh.mesh import Mesh, EuclideanMesh2D
+from ocsmesh.mesh.base import BaseMesh
 from ocsmesh.raster import Raster, get_iter_windows
 from ocsmesh.features.contour import Contour
 from ocsmesh.features.patch import Patch
 from ocsmesh.features.linefeature import LineFeature
 from ocsmesh.features.channel import Channel
 from ocsmesh.features.constraint import (
-    TopoConstConstraint, TopoFuncConstraint, CourantNumConstraint
+    TopoConstConstraint,
+    TopoFuncConstraint,
+    CourantNumConstraint,
+    RegionConstraint,
 )
 
 CanCreateSingleHfun = Union[Raster, EuclideanMesh2D]
@@ -65,6 +69,12 @@ CanCreateHfun = Union[CanCreateSingleHfun, CanCreateMultipleHfun]
 
 SizeFuncList = Sequence[Union[HfunRaster, HfunMesh]]
 
+RASTER_CONSTR = (
+    TopoConstConstraint,
+    TopoFuncConstraint,
+    CourantNumConstraint,
+)
+
 _logger = logging.getLogger(__name__)
 
 class _RefinementContourInfoCollector:
@@ -533,6 +543,22 @@ class _ConstraintInfoCollector:
             yield defn
 
 
+    def apply(self, hfun_list, per_hfun=True):
+        for in_idx, hfun in enumerate(hfun_list):
+            is_raster = isinstance(hfun, HfunRaster)
+            constraint_list = []
+            for src_idx, constraint_defn in self:
+                if per_hfun and src_idx is not None and in_idx not in src_idx:
+                    continue
+
+                if isinstance(constraint_defn, RASTER_CONSTR) and not is_raster:
+                    continue
+
+                constraint_list.append(constraint_defn)
+
+            if constraint_list:
+                hfun.apply_constraints(constraint_list)
+
 
 class HfunCollector(BaseHfun):
     """Define size function based on multiple inputs of different types
@@ -558,6 +584,8 @@ class HfunCollector(BaseHfun):
         contraction rate `rate` specified. This constraint is only
         applied on rasters with specified indices. The index is w.r.t
         the full input list for collector object creation.
+    add_region_constraint(...)
+        Add a constraint on size based on a specified region
     add_patch(...)
         Add a region of fixed size refinement with optional expansion
         rate for points outside the region to achieve smooth size
@@ -722,7 +750,8 @@ class HfunCollector(BaseHfun):
             elif isinstance(in_item, EuclideanMesh2D):
                 hfun = HfunMesh(in_item)
 
-            elif isinstance(in_item, str):
+            elif isinstance(in_item, (str, Path)):
+                in_item = str(in_item)
                 if in_item.endswith('.tif'):
                     raster = Raster(in_item)
                     if self._base_shape:
@@ -847,7 +876,7 @@ class HfunCollector(BaseHfun):
             Rate of relaxation of constraint outside the region defined
             by `lower_bound` and `upper_bound`.
         source_index : int or list of ints or None, default=None
-            The index of raster entries from the input list argument
+            The index of entries from the input list argument
             of the constructor of collector size function. If `None`
             all input rasters are used.
 
@@ -904,7 +933,7 @@ class HfunCollector(BaseHfun):
             Rate of relaxation of constraint outside the region defined
             by `lower_bound` and `upper_bound`.
         source_index : int or list of ints or None, default=None
-            The index of raster entries from the input list argument
+            The index of entries from the input list argument
             of the constructor of collector size function. If `None`
             all input rasters are used.
 
@@ -946,7 +975,7 @@ class HfunCollector(BaseHfun):
             Free surface elevation (:math:`meters`) from the reference
             (i.e. wave height)
         source_index : int or list of ints or None, default=None
-            The index of raster entries from the input list argument
+            The index of entries from the input list argument
             of the constructor of collector size function. If `None`
             all input rasters are used.
 
@@ -986,6 +1015,56 @@ class HfunCollector(BaseHfun):
             )
 
 
+    def add_region_constraint(
+            self,
+            value: Union[float, npt.NDArray[np.float32]],
+            shape: Union[Polygon, MultiPolygon],
+            crs: Union[CRS, str] = None,
+            value_type: Literal['min', 'max'] = 'min',
+            rate: float = 0.01,
+            source_index: Union[List[int], int, None] = None
+            ) -> None:
+
+        """Add a value contraint for the points in specified region
+
+        Add a fixed-value constraint to the region of the size function
+        specified by the input shape.  Optionally a `rate` can be
+        specified to relax the constraint gradually outside the bounds.
+
+        Parameters
+        ----------
+        value : float or array-like
+            A single fixed value to be used for
+            mesh size if condition is not met based on `value_type`.
+        shape: MultiPolygon or Polygon
+            Region of specified constraint
+        crs: CRS or None, default=None
+            The CRS of the input shape
+        value_type : {'min', 'max'}, default='min'
+            Type of contraint. If 'min', it means the mesh size
+            should not be smaller than the specified `value` at each
+            point.
+        rate : float, default=0.01
+            Rate of relaxation of constraint outside the specified region
+        source_index: int or list of ints or None, default=None
+            The index of input hfun source. If `None` all inputs
+            are used.
+
+        Returns
+        -------
+        None
+        """
+        self._constraint_info_coll.add(
+            source_index,
+            RegionConstraint(
+                value=value,
+                shape=shape,
+                crs=crs,
+                value_type=value_type,
+                rate=rate
+            )
+        )
+
 
     def add_contour(
             self,
@@ -1283,7 +1362,7 @@ class HfunCollector(BaseHfun):
     def add_feature(
             self,
             shape: Union[MultiLineString, LineString, None] = None,
-            line_defn: Optional[LineString] = None,
+            line_defn: Optional[LineFeature] = None,
             shapefile: Union[None, str, Path] = None,
             expansion_rate: float = 0.01,
             target_size: Optional[float] = None,
@@ -1359,7 +1438,7 @@ class HfunCollector(BaseHfun):
             If the any of the inputs are not supported
         """
 
-        valid_types = (str, Raster, Mesh, HfunRaster, HfunMesh)
+        valid_types = (str, Path, Raster, BaseMesh, HfunRaster, HfunMesh)
         if not all(isinstance(item, valid_types) for item in input_list):
             raise TypeError(
                 f'Input list items must be of type'
@@ -1417,19 +1496,7 @@ class HfunCollector(BaseHfun):
             raise NotImplementedError(
                 "This function does not suuport fast hfun method")
 
-        raster_hfun_list = [
-            i for i in self._hfun_list if isinstance(i, HfunRaster)]
-
-        for in_idx, hfun in enumerate(raster_hfun_list):
-            constraint_list = []
-            for src_idx, constraint_defn in self._constraint_info_coll:
-                if src_idx is not None and in_idx not in src_idx:
-                    continue
-
-                constraint_list.append(constraint_defn)
-
-            if constraint_list:
-                hfun.apply_constraints(constraint_list)
+        self._constraint_info_coll.apply(self._hfun_list)
 
 
     def _apply_contours(self, apply_to: Optional[SizeFuncList] = None) -> None:
@@ -2124,13 +2191,8 @@ class HfunCollector(BaseHfun):
         _apply_constraints
         """
 
-        constraint_list = []
-        for src_idx, constraint_defn in self._constraint_info_coll:
-            # TODO: Account for source index
-            constraint_list.append(constraint_defn)
-
-        if constraint_list:
-            big_hfun.apply_constraints(constraint_list)
+        # TODO: Account for source index
+        self._constraint_info_coll.apply([big_hfun], per_hfun=False)
 
 
     def _get_hfun_composite_fast(self, big_hfun):
