@@ -8,7 +8,7 @@ from multiprocessing import cpu_count, Pool
 import operator
 import tempfile
 from time import time
-from typing import Union, List, Callable, Any, Optional, Iterable, Tuple
+from typing import Union, List, Callable, Optional, Iterable, Tuple
 from contextlib import ExitStack
 import warnings
 try:
@@ -37,7 +37,9 @@ from ocsmesh.features.constraint import (
     Constraint,
     TopoConstConstraint,
     TopoFuncConstraint,
-    CourantNumConstraint
+    CourantNumConstraint,
+    RegionConstraint,
+    apply_constraints_wrap
 )
 from ocsmesh import utils
 
@@ -48,35 +50,6 @@ warnings.filterwarnings(
 _logger = logging.getLogger(__name__)
 
 
-def _apply_constraints(method: Callable[..., Any]) -> Callable[..., Any]:
-    """Decorator for (re)applying constraint after updating hfun spec
-
-    This function is a deocrator that takes in a callable and assumes
-    the callable is a method whose first argument is an object. In the
-    wrapper function, after the wrapped method is called the
-    `apply_added_constraints` method is called on the object and
-    then the return value of the wrapped method is returned to the
-    caller.
-
-    Parameters
-    ----------
-    method : callable
-        Method to be wrapped so that the constraints are automatically
-        re-applied after method is executed.
-
-    Returns
-    -------
-    callable
-        The wrapped method
-    """
-
-    def wrapped(obj, *args, **kwargs):
-        rv = method(obj, *args, **kwargs)
-        obj.apply_added_constraints()
-        return rv
-    return wrapped
-
-
 class HfunInputRaster:
     """Descriptor class for holding reference to the input raster"""
 
@@ -145,6 +118,8 @@ class HfunRaster(BaseHfun, Raster):
         contraction rate `rate` specified.
     add_courant_num_constraint(...)
         Add constraint based on approximated Courant number
+    add_region_constraint(...)
+        Add a constraint on size based on a specified region
     add_patch(...)
         Add a region of fixed size refinement with optional expansion
         rate for points outside the region to achieve smooth size
@@ -574,6 +549,8 @@ class HfunRaster(BaseHfun, Raster):
                 if utm_crs is not None:
                     xy = self.get_xy_memcache(window, utm_crs)
                 else:
+                    # Technically it means that crs is not geographic!
+                    utm_crs = self.crs
                     xy = self.get_xy(window)
 
 
@@ -581,7 +558,7 @@ class HfunRaster(BaseHfun, Raster):
                 _logger.debug(f'Processing window {i+1}/{tot}.')
                 for constraint in constraint_list:
                     hfun_values = constraint.apply(
-                            rast_values, hfun_values, locations=xy)
+                            rast_values, hfun_values, locations=xy, crs=utm_crs)
 
                 # Apply global constraints
                 if self.hmin is not None:
@@ -594,7 +571,7 @@ class HfunRaster(BaseHfun, Raster):
                 gc.collect()
 
 
-    @_apply_constraints
+    @apply_constraints_wrap
     def add_topo_bound_constraint(
             self,
             value: Union[float, npt.NDArray[np.float32]],
@@ -648,7 +625,7 @@ class HfunRaster(BaseHfun, Raster):
             value, upper_bound, lower_bound, value_type, rate))
 
 
-    @_apply_constraints
+    @apply_constraints_wrap
     def add_topo_func_constraint(
             self,
             func: Callable[[npt.NDArray[np.float32]], npt.NDArray[np.float32]]
@@ -702,7 +679,7 @@ class HfunRaster(BaseHfun, Raster):
         self._constraints.append(TopoFuncConstraint(
             func, upper_bound, lower_bound, value_type, rate))
 
-    @_apply_constraints
+    @apply_constraints_wrap
     def add_courant_num_constraint(
             self,
             upper_bound: float = 0.9,
@@ -761,7 +738,71 @@ class HfunRaster(BaseHfun, Raster):
             )
 
 
-    @_apply_constraints
+    @apply_constraints_wrap
+    def add_region_constraint(
+            self,
+            value: Union[float, npt.NDArray[np.float32]],
+            shape: Union[Polygon, MultiPolygon],
+            crs: Union[CRS, str] = None,
+            value_type: Literal['min', 'max'] = 'min',
+            rate: float = 0.01
+            ) -> None:
+        """Add a value contraint for the points in specified region
+
+        Add a fixed-value constraint to the region of the size function
+        specified by the input shape.  Optionally a `rate` can be
+        specified to relax the constraint gradually outside the bounds.
+
+        Parameters
+        ----------
+        value : float or array-like
+            A single fixed value to be used for
+            mesh size if condition is not met based on `value_type`.
+        shape: MultiPolygon or Polygon
+            Region of specified constraint
+        crs: CRS or None, default=None
+            The CRS of the input shape
+        value_type : {'min', 'max'}, default='min'
+            Type of contraint. If 'min', it means the mesh size
+            should not be smaller than the specified `value` at each
+            point.
+        rate : float, default=0.01
+            Rate of relaxation of constraint outside the specified region
+
+        Returns
+        -------
+        None
+
+        See Also
+        --------
+        add_topo_bound_constraint :
+            Add fixed-value or fixed-matrix constraint.
+        add_topo_func_constraint :
+            Addint constraint based on function of topography
+        add_courant_num_constraint :
+            Add constraint based on approximated Courant number
+        """
+
+        if crs is None:
+            warnings.warn(
+                "No CRS is provided for the contraint shape."
+                + " Assuming CRS is the same as size function's raster",
+                category=UserWarning
+            )
+            crs = self.crs
+
+        self._constraints.append(
+            RegionConstraint(
+                value=value,
+                shape=shape,
+                crs=crs,
+                value_type=value_type,
+                rate=rate
+            )
+        )
+
+
+    @apply_constraints_wrap
     def add_patch(
             self,
             multipolygon: Union[MultiPolygon, Polygon],
@@ -836,9 +877,9 @@ class HfunRaster(BaseHfun, Raster):
 
         # For expansion_rate
         if expansion_rate is not None:
-            exteriors = [ply.exterior for ply in multipolygon]
+            exteriors = [ply.exterior for ply in multipolygon.geoms]
             interiors = [
-                inter for ply in multipolygon for inter in ply.interiors]
+                inter for ply in multipolygon.geoms for inter in ply.interiors]
 
             features = MultiLineString([*exteriors, *interiors])
             # pylint: disable=E1123, E1125
@@ -862,7 +903,7 @@ class HfunRaster(BaseHfun, Raster):
                 start = time()
                 try:
                     mask, _, _ = rasterio.mask.raster_geometry_mask(
-                        self.src, multipolygon,
+                        self.src, multipolygon.geoms,
                         all_touched=True, invert=True)
                     mask = mask[rasterio.windows.window_index(window)]
 
@@ -894,7 +935,7 @@ class HfunRaster(BaseHfun, Raster):
                 _logger.info(f'Write array to file took {time()-start}.')
 
 
-    @_apply_constraints
+    @apply_constraints_wrap
     def add_contour(
             self,
             level: Union[List[float], float],
@@ -974,7 +1015,7 @@ class HfunRaster(BaseHfun, Raster):
                 contours, expansion_rate, target_size,
                 nprocs=nprocs)
 
-    @_apply_constraints
+    @apply_constraints_wrap
     def add_channel(
             self,
             level: float = 0,
@@ -1042,7 +1083,7 @@ class HfunRaster(BaseHfun, Raster):
 
 
 
-    @_apply_constraints
+    @apply_constraints_wrap
     @utils.add_pool_args
     def add_feature(
             self,
@@ -1269,7 +1310,7 @@ class HfunRaster(BaseHfun, Raster):
         return np.memmap(tmpfile, dtype='float32', mode='r',
                          shape=((window.width*window.height), 2))[:]
 
-    @_apply_constraints
+    @apply_constraints_wrap
     def add_subtidal_flow_limiter(
             self,
             hmin: Optional[float] = None,
@@ -1392,7 +1433,7 @@ class HfunRaster(BaseHfun, Raster):
                     self.dtype(1))
                 dst.write_band(1, hfun_values, window=window)
 
-    @_apply_constraints
+    @apply_constraints_wrap
     def add_constant_value(
             self,
             value: float,
