@@ -12,6 +12,7 @@ import matplotlib.pyplot as plt  # type: ignore[import]
 from matplotlib.tri import Triangulation  # type: ignore[import]
 import numpy as np  # type: ignore[import]
 import numpy.typing as npt
+import rasterio as rio
 from pyproj import CRS, Transformer  # type: ignore[import]
 from scipy.interpolate import (  # type: ignore[import]
     RectBivariateSpline, griddata)
@@ -25,7 +26,13 @@ import geopandas as gpd
 import utm
 
 
+# TODO: Remove one of these two constants
 ELEM_2D_TYPES = ['tria3', 'quad4', 'hexa8']
+MESH_TYPES = {
+    'tria3': 'TRIA3_t',
+    'quad4': 'QUAD4_t',
+    'hexa8': 'HEXA8_t'
+}
 
 def must_be_euclidean_mesh(func):
     def decorator(mesh, *args, **kwargs):
@@ -48,27 +55,77 @@ def mesh_to_tri(mesh):
 
 def cleanup_isolates(mesh):
 
-    # For triangle only (TODO: add support for other types)
-    node_indexes = np.arange(mesh.vert2['coord'].shape[0])
-    used_indexes = np.unique(mesh.tria3['index'])
-    vert2_idxs = np.where(
-        np.isin(node_indexes, used_indexes, assume_unique=True))[0]
+    used_old_idx = np.array([], dtype='int64')
+    for etype in ELEM_2D_TYPES:
+        elem_idx = getattr(mesh, etype)['index'].flatten()
+        used_old_idx = np.hstack((used_old_idx, elem_idx))
+    used_old_idx = np.unique(used_old_idx)
 
-    # Since tria simplex refers to node index which always starts from
-    # 0 after removing isolate nodes we can use the map approach
-    tria3 = mesh.tria3['index'].flatten()
-    renum = {old: new for new, old in enumerate(np.unique(tria3))}
-    tria3 = np.array([renum[i] for i in tria3])
-    tria3 = tria3.reshape(mesh.tria3['index'].shape)
-
-    mesh.vert2 = mesh.vert2.take(vert2_idxs, axis=0)
+    # update vert2 and value
+    mesh.vert2 = mesh.vert2.take(used_old_idx, axis=0)
     if len(mesh.value) > 0:
-        mesh.value = mesh.value.take(vert2_idxs, axis=0)
-    mesh.tria3 = np.asarray(
-        [(tuple(indices), mesh.tria3['IDtag'][i])
-         for i, indices in enumerate(tria3)],
-        dtype=jigsaw_msh_t.TRIA3_t)
+        mesh.value = mesh.value.take(used_old_idx, axis=0)
 
+    renum = {old: new for new, old in enumerate(np.unique(used_old_idx))}
+    for etype in ELEM_2D_TYPES:
+        elem_idx = getattr(mesh, etype)['index']
+        elem_new_idx = np.array([renum[i] for i in elem_idx.flatten()])
+        elem_new_idx = elem_new_idx.reshape(elem_idx.shape)
+        # TODO: Keep IDTag?
+        setattr(mesh, etype, np.array(
+                [(idx, 0) for idx in elem_new_idx],
+                dtype=getattr(
+                    jigsawpy.jigsaw_msh_t, f'{etype.upper()}_t')))
+
+
+def cleanup_duplicates(mesh):
+
+    """Cleanup duplicate nodes and elements
+
+    Notes
+    -----
+    Elements and nodes are duplicate if they fully overlapping (not
+    partially)
+    """
+
+    _, cooidx, coorev = np.unique(
+        mesh.vert2['coord'],
+        axis=0,
+        return_index=True,
+        return_inverse=True
+    )
+    nd_map = {e: i for e, i in enumerate(coorev)}
+    mesh.vert2 = mesh.vert2.take(cooidx, axis=0)
+
+    for etype, otype in MESH_TYPES.items():
+        cnn = getattr(mesh, etype)['index']
+
+        n_node = cnn.shape[1]
+
+        cnn_renu = np.array(
+            [nd_map[i] for i in cnn.flatten()]
+        ).reshape(-1, n_node)
+
+        _, cnnidx = np.unique(
+            np.sort(cnn_renu, axis=1),
+            axis=0,
+            return_index=True
+        )
+        mask = np.zeros(len(cnn_renu), dtype=bool)
+        mask[cnnidx] = True
+        adj_cnn = cnn_renu[mask]
+
+        setattr(
+            mesh,
+            etype,
+            np.array(
+                [(idx, 0) for idx in adj_cnn],
+                dtype=getattr(jigsaw_msh_t, otype)
+            )
+        )
+
+    if len(mesh.value) > 0:
+        mesh.value = mesh.value.take(sorted(cooidx), axis=0)
 
 def put_edge2(mesh):
     tri = Triangulation(
@@ -184,8 +241,9 @@ def get_mesh_polygons(mesh):
 
             res_gdf = polys_gdf[polys_gdf.intersects(pnt)]
             if len(res_gdf) == 0:
-                # How is this possible?!
-                pnts = MultiPoint([*(pnts.geoms[:idx]), *(pnts.geoms[idx + 1:])])
+                # How is this possible for the remaining points not to
+                # intersect with the remaining polys?!
+                pnts = MultiPoint([pt for i, pt in enumerate(pnts.geoms) if i != idx])
                 if pnts.is_empty:
                     break
 
@@ -321,31 +379,32 @@ def _sieve_by_mask(mesh, sieve_mask):
     lone_elem_verts = get_lone_element_verts(mesh)
     vert2_mask[lone_elem_verts] = True
 
-    # Mask out elements containing the unwanted nodes.
-    tria3_mask = np.any(vert2_mask[mesh.tria3['index']], axis=1)
+    used_old_idx = np.array([], dtype='int64')
+    filter_dict = {}
+    for etype in ELEM_2D_TYPES:
+        elem_idx = getattr(mesh, etype)['index']
+        elem_mask = np.any(vert2_mask[elem_idx], axis=1)
 
-    # Tria and node removal and renumbering indexes ...
-    tria3_id_tag = mesh.tria3['IDtag'].take(np.where(~tria3_mask)[0])
-    tria3_index = mesh.tria3['index'][~tria3_mask, :].flatten()
-    used_indexes = np.unique(tria3_index)
-    node_indexes = np.arange(mesh.vert2['coord'].shape[0])
-    renum = {old: new for new, old in enumerate(np.unique(tria3_index))}
-    tria3_index = np.array([renum[i] for i in tria3_index])
-    tria3_index = tria3_index.reshape((tria3_id_tag.shape[0], 3))
-    vert2_idxs = np.where(np.isin(node_indexes, used_indexes))[0]
+        # Tria and node removal and renumbering indexes ...
+        elem_keep_idx = elem_idx[~elem_mask, :].flatten()
+        used_old_idx = np.hstack((used_old_idx, elem_keep_idx))
+        filter_dict[etype] = [elem_keep_idx, elem_idx.shape[1]]
+    used_old_idx = np.unique(used_old_idx)
 
-    # update vert2
-    mesh.vert2 = mesh.vert2.take(vert2_idxs, axis=0)
-
-    # update value
+    # update vert2 and value
+    mesh.vert2 = mesh.vert2.take(used_old_idx, axis=0)
     if len(mesh.value) > 0:
-        mesh.value = mesh.value.take(vert2_idxs, axis=0)
+        mesh.value = mesh.value.take(used_old_idx, axis=0)
 
-    # update tria3
-    mesh.tria3 = np.array(
-        [(tuple(indices), tria3_id_tag[i])
-         for i, indices in enumerate(tria3_index)],
-        dtype=jigsaw_msh_t.TRIA3_t)
+    renum = {old: new for new, old in enumerate(np.unique(used_old_idx))}
+    for etype, (elem_keep_idx, topo) in filter_dict.items():
+        elem_new_idx = np.array([renum[i] for i in elem_keep_idx])
+        elem_new_idx = elem_new_idx.reshape(-1, topo)
+        # TODO: Keep IDTag?
+        setattr(mesh, etype, np.array(
+                [(idx, 0) for idx in elem_new_idx],
+                dtype=getattr(
+                    jigsawpy.jigsaw_msh_t, f'{etype.upper()}_t')))
 
 
 def finalize_mesh(mesh, sieve_area=None):
@@ -374,6 +433,7 @@ def finalize_mesh(mesh, sieve_area=None):
             break
 
     cleanup_isolates(mesh)
+    cleanup_duplicates(mesh)
     put_id_tags(mesh)
 
 
@@ -440,49 +500,7 @@ def sieve(mesh, area=None):
         path = Path(multipolygon[idx].exterior.coords, closed=True)
         vert2_mask = vert2_mask | path.contains_points(mesh.vert2['coord'])
 
-    # select any connected nodes; these ones are missed by
-    # path.contains_point() because they are at the path edges.
-    _idxs = np.where(vert2_mask)[0]
-    conn_verts = get_surrounding_elem_verts(mesh, _idxs)
-    vert2_mask[conn_verts] = True
-
-    # Also, there might be some dangling triangles without neighbors,
-    # which are also missed by path.contains_point()
-    lone_elem_verts = get_lone_element_verts(mesh)
-    vert2_mask[lone_elem_verts] = True
-
-
-    # Mask out elements containing the unwanted nodes.
-    tria3_mask = np.any(vert2_mask[mesh.tria3['index']], axis=1)
-
-    # Renumber indexes ...
-    # isolated node removal does not require elimination of triangles from
-    # the table, therefore the length of the indexes is constant.
-    # We must simply renumber the tria3 indexes to match the new node indexes.
-    # Essentially subtract one, but going from the bottom of the index table
-    # to the top.
-    used_indexes = np.unique(mesh.tria3['index'])
-    node_indexes = np.arange(mesh.vert2['coord'].shape[0])
-    tria3_idxs = np.where(~np.isin(node_indexes, used_indexes))[0]
-    tria3_id_tag = mesh.tria3['IDtag'].take(np.where(~tria3_mask)[0])
-    tria3_index = mesh.tria3['index'][~tria3_mask, :].flatten()
-    for idx in reversed(tria3_idxs):
-        tria3_index[np.where(tria3_index >= idx)] -= 1
-    tria3_index = tria3_index.reshape((tria3_id_tag.shape[0], 3))
-    vert2_idxs = np.where(np.isin(node_indexes, used_indexes))[0]
-
-    # update vert2
-    mesh.vert2 = mesh.vert2.take(vert2_idxs, axis=0)
-
-    # update value
-    if len(mesh.value) > 0:
-        mesh.value = mesh.value.take(vert2_idxs, axis=0)
-
-    # update tria3
-    mesh.tria3 = np.array(
-        [(tuple(indices), tria3_id_tag[i])
-         for i, indices in enumerate(tria3_index)],
-        dtype=jigsaw_msh_t.TRIA3_t)
+    return _sieve_by_mask(mesh, vert2_mask)
 
 
 def sort_edges(edges):
@@ -625,6 +643,7 @@ def inner_ring_collection(mesh):
 def get_multipolygon_from_pathplot(ax):
     # extract linear_rings from plot
     linear_ring_collection = []
+    multipolygon = None
     for path_collection in ax.collections:
         for path in path_collection.get_paths():
             polygons = path.to_polygons(closed_only=True)
@@ -651,7 +670,7 @@ def get_multipolygon_from_pathplot(ax):
             polygon_collection.append(Polygon(outer_ring, inner_rings))
 
         multipolygon = MultiPolygon(polygon_collection)
-    else:
+    elif len(linear_ring_collection) != 0:
         multipolygon = MultiPolygon(
             [Polygon(linear_ring_collection.pop())])
     return multipolygon
@@ -788,7 +807,7 @@ def get_verts_in_shape(
     # We need point indices in the shapes, not the shape indices
     # query bulk returns all combination of intersections in case
     # input shape results in multi-row series
-    in_shp_idx = pt_series.sindex.query_bulk(
+    in_shp_idx = pt_series.sindex.query(
             shp_series, predicate="intersects")[1]
 
     in_shp_idx = select_adjacent(mesh, in_shp_idx, num_layers=num_adjacent)
@@ -806,13 +825,8 @@ def select_adjacent(mesh, in_indices, num_layers):
             coord = mesh.vert2['coord']
 
             # TODO: What about edge2
-            mesh_types = {
-                'tria3': 'TRIA3_t',
-                'quad4': 'QUAD4_t',
-                'hexa8': 'HEXA8_t'
-            }
             elm_dict = {
-                key: getattr(mesh, key)['index'] for key in mesh_types}
+                key: getattr(mesh, key)['index'] for key in MESH_TYPES}
 
             mark_func = np.any
 
@@ -996,13 +1010,8 @@ def clip_mesh_by_vertex(
         coord = mesh.vert2['coord']
 
         # TODO: What about edge2 if in_place?
-        mesh_types = {
-            'tria3': 'TRIA3_t',
-            'quad4': 'QUAD4_t',
-            'hexa8': 'HEXA8_t'
-        }
         elm_dict = {
-            key: getattr(mesh, key)['index'] for key in mesh_types}
+            key: getattr(mesh, key)['index'] for key in MESH_TYPES}
 
         # Whether elements that include "in"-vertices can be created
         # using vertices other than "in"-vertices
@@ -1065,7 +1074,7 @@ def clip_mesh_by_vertex(
             [(coo, 0) for coo in new_coord],
             dtype=jigsaw_msh_t.VERT2_t)
 
-        for key, elem_type in mesh_types.items():
+        for key, elem_type in MESH_TYPES.items():
             setattr(
                 mesh_out,
                 key,
@@ -1642,14 +1651,8 @@ def merge_msh_t(
 
     dst_crs = CRS.from_user_input(out_crs)
 
-    mesh_types = {
-        'tria3': 'TRIA3_t',
-        'quad4': 'QUAD4_t',
-        'hexa8': 'HEXA8_t'
-    }
-
     coord = []
-    elems = {k: [] for k in mesh_types}
+    elems = {k: [] for k in MESH_TYPES}
     value = []
     offset = 0
 
@@ -1681,7 +1684,7 @@ def merge_msh_t(
         mesh_shape_list.append(mesh_shape)
 
 
-        for k in mesh_types:
+        for k in MESH_TYPES:
             cnn = getattr(mesh, k)
             elems[k].append(cnn['index'] + offset)
         coord.append(mesh.vert2['coord'])
@@ -1698,7 +1701,7 @@ def merge_msh_t(
     composite_mesh.value = np.array(
             np.vstack(value),
             dtype=jigsaw_msh_t.REALS_t)
-    for k, v in mesh_types.items():
+    for k, v in MESH_TYPES.items():
         setattr(composite_mesh, k, np.array(
             [(cnn, 0) for cnn in np.vstack(elems[k])],
             dtype=getattr(jigsaw_msh_t, v)))
@@ -2034,3 +2037,154 @@ def approximate_courant_number_for_depth(
     # For overland where h < nu the characteristic velocity is 2 * sqrt(g*h)
     characteristic_velocity_magnitude[~depth_mask] = 2 * particle_velocity[~depth_mask]
     return characteristic_velocity_magnitude * timestep / element_size
+
+
+def create_rectangle_mesh(
+    nx,
+    ny,
+    holes,
+    x_extent=None,
+    y_extent=None,
+    quads=None,
+):
+    # pylint: disable=W1401
+    """
+    Note:
+        x = x-index
+        y = y-index
+
+        holes or quads count starting at 1 from bottom corner square
+
+        node-index(node-id)
+
+              25(26)             29(30)
+          5     *---*---*---*---*
+                | \ | \ | \ | \ |
+          4     *---*---*---*---*
+                | \ | \ | \ | \ |
+          3     *---*---*---*---*
+                | \ |   | \ | \ |
+          2     *---*---*---*---*
+                | \ | \ | # | \ |
+          1     *---*---*---*---*
+                | \ | \ | \ | \ |
+          0     *---*---*---*---*
+              0(1)               4(5)
+
+                0   1   2   3   4
+    """
+
+    if x_extent is None:
+        x_range = range(nx)
+    else:
+        x_range = np.linspace(x_extent[0], x_extent[1], nx)
+
+    if y_extent is None:
+        y_range = range(ny)
+    else:
+        y_range = np.linspace(y_extent[0], y_extent[1], ny)
+
+    if quads is None:
+        quads = []
+
+    X, Y = np.meshgrid(x_range, y_range)
+    verts = np.array(list(zip(X.ravel(), Y.ravel())))
+    tria3 = []
+    quad4 = []
+    for j in range(ny - 1):
+        for i in range(nx - 1):
+            is_quad = (i + 1) + ((nx-1) * j) in quads
+            is_hole = (i + 1) + ((nx-1) * j) in holes
+            if is_hole:
+                continue
+            if is_quad:
+                quad4.append([
+                    j * nx + i,
+                    j * nx + (i + 1),
+                    (j + 1) * nx + (i + 1),
+                    (j + 1) * nx + i
+                ])
+            else: # is tria
+                tria3.append([j * nx + i, j * nx + (i + 1), (j + 1) * nx + i])
+                tria3.append([j * nx + (i + 1), (j + 1) * nx + (i + 1), (j + 1) * nx + i])
+
+
+
+    # NOTE: Everywhere is above 0 (auto: land) unless modified later
+    vals = np.ones((len(verts), 1)) * 10
+
+    mesh_msht = msht_from_numpy(
+        coordinates=verts,
+        triangles=tria3,
+        quadrilaterals=quad4 if len(quad4) > 0 else None,
+        crs=None
+    )
+
+    # Drop unused verts (e.g. 4 connected holes)
+    cleanup_isolates(mesh_msht)
+    mesh_msht.value = np.array(
+        vals, dtype=jigsaw_msh_t.REALS_t
+    )
+
+    return mesh_msht
+
+
+def raster_from_numpy(
+    filename,
+    data,
+    mgrid,
+    crs=CRS.from_epsg(4326)
+) -> None:
+    x = mgrid[0][:, 0]
+    y = mgrid[1][0, :]
+    res_x = (x[-1] - x[0]) / data.shape[1]
+    res_y = (y[-1] - y[0]) / data.shape[0]
+    # TODO: Mistake in transformation if x and y extent are different?
+    transform = rio.transform.Affine.translation(
+        x[0], y[0]
+    ) * rio.transform.Affine.scale(res_x, res_y)
+    if not isinstance(crs, CRS):
+        crs = CRS.from_user_input(crs)
+
+    with rio.open(
+        filename,
+        'w',
+        driver='GTiff',
+        height=data.shape[0],
+        width=data.shape[1],
+        count=1,
+        dtype=data.dtype,
+        crs=crs,
+        transform=transform,
+    ) as dst:
+        dst.write(data, 1)
+
+
+def msht_from_numpy(
+    coordinates,
+    triangles,
+    quadrilaterals=None,
+    crs=CRS.from_epsg(4326)
+):
+    mesh = jigsaw_msh_t()
+    mesh.mshID = 'euclidean-mesh'
+    mesh.ndims = +2
+    if crs is not None:
+        if not isinstance(crs, CRS):
+            crs = CRS.from_user_input(crs)
+        mesh.crs = crs
+    mesh.vert2 = np.array(
+        [(coord, 0) for coord in coordinates],
+        dtype=jigsaw_msh_t.VERT2_t
+        )
+    mesh.tria3 = np.array(
+        [(index, 0) for index in triangles],
+        dtype=jigsaw_msh_t.TRIA3_t
+        )
+    if quadrilaterals is not None:
+        mesh.quad4 = np.array(
+            [(index, 0) for index in quadrilaterals],
+            dtype=jigsaw_msh_t.QUAD4_t
+            )
+
+    return mesh
